#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tenable integration for RegScale CLI"""

# standard python imports
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    import pandas as pd

import collections
import json
import os
import re
import tempfile
import time
import uuid
import warnings
from datetime import datetime, timedelta
from itertools import groupby
from pathlib import Path
from typing import Dict, List, Set, Tuple, Union
from urllib.parse import urljoin

import click
import matplotlib.pyplot as plt
import requests
from requests.exceptions import RequestException
from rich.console import Console
from rich.pretty import pprint
from rich.progress import track
from tenable.io import TenableIO
from tenable.sc import TenableSC

from regscale import __version__
from regscale.core.app.api import Api
from regscale.core.app.application import Application
from regscale.core.app.logz import create_logger
from regscale.core.app.utils.app_utils import (
    check_file_path,
    check_license,
    create_progress_object,
    epoch_to_datetime,
    error_and_exit,
    format_dict_to_html,
    get_current_datetime,
    regscale_string_to_epoch,
    save_data_to,
)
from regscale.core.app.utils.nessus_utils import (
    determine_available_space,
    get_cpe_file,
    get_min_cvss_score,
    validate_nessus_severity,
)
from regscale.models.app_models.click import file_types, save_output_to
from regscale.models.integration_models.tenable import AssetCheck, TenableAsset, TenableIOAsset
from regscale.models.regscale_models import ControlImplementation
from regscale.models.regscale_models.asset import Asset
from regscale.models.regscale_models.issue import Issue
from regscale.models.regscale_models.scan_history import ScanHistory
from regscale.validation.address import validate_mac_address
from regscale.validation.record import validate_regscale_object

console = Console()

logger = create_logger("rich")
REGSCALE_INC = "RegScale, Inc."
REGSCALE_CLI = "RegScale CLI"

FULLY_IMPLEMENTED = "Fully Implemented"
NOT_IMPLEMENTED = "Not Implemented"
IN_REMEDIATION = "In Remediation"

DONE_MSG = "Done!"


#####################################################################################################
#
# Tenable.sc Documentation: https://docs.tenable.com/tenablesc/api/index.htm
# pyTenable GitHub repo: https://github.com/tenable/pyTenable
# Python tenable.sc documentation: https://pytenable.readthedocs.io/en/stable/api/sc/index.html
#
#####################################################################################################


# Create group to handle OSCAL processing
@click.group()
def tenable():
    """Performs actions on the Tenable APIs."""


@tenable.group(help="[BETA] Performs actions on the Tenable.io API.")
def io():
    """Performs actions on the Tenable.io API."""


@tenable.group(help="[BETA] Performs actions on the Tenable.sc API.")
def sc():
    """Performs actions on the Tenable.sc API."""


@tenable.group(help="[BETA] Import Nessus scans and assets to RegScale.")
def nessus():
    """Performs actions on the Tenable.sc API."""


@nessus.command(name="import_nessus")
@click.option(
    "--folder_path",
    prompt="Enter the folder path of the Nessus files to process",
    help="RegScale will load the Nessus Scans",
    type=click.Path(exists=True),
)
@click.option(
    "--regscale_ssp_id",
    type=click.INT,
    help="The ID number from RegScale of the System Security Plan",
    prompt="Enter RegScale System Security Plan ID",
    required=True,
)
def import_nessus(folder_path: click.Path, regscale_ssp_id: click.INT):
    """Import Nessus scans, vulnerabilities and assets to RegScale."""
    from regscale.models.integration_models.nessus import NessusReport

    NessusReport.process_nessus_files(folder_path=folder_path, regscale_ssp_id=regscale_ssp_id)


@nessus.command(name="update_cpe_dictionary")
def update_cpe_dictionary():
    """
    Manually update the CPE 2.2 dictionary from NIST.
    """
    get_cpe_file(download=True)


@sc.command(name="export_scans")
@save_output_to()
@file_types([".json", ".csv", ".xlsx"])
def export_scans(save_output_to: Path, file_type: str):
    """Export scans from Tenable Host to a .json, .csv or .xlsx file."""
    # get the scan results
    results = get_usable_scan_list()

    # check if file path exists
    check_file_path(save_output_to)

    # set the file name
    file_name = f"tenable_scans_{get_current_datetime('%m%d%Y')}"

    # save the data as the selected file by the user
    save_data_to(
        file=Path(f"{save_output_to}/{file_name}{file_type}"),
        data=results,
    )


def get_usable_scan_list() -> list:
    """
    Usable Scans from Tenable Host

    :return: List of scans from Tenable
    :rtype: list
    """
    results = []
    try:
        client = gen_client()
        results = client.scans.list()["usable"]
    except Exception as ex:
        logger.error(ex)
    return results


def get_detailed_scans(scan_list: list = None) -> list:
    """
    Generate list of detailed scans (Warning: this action could take 20 minutes or more to complete)

    :param list scan_list: List of scans from Tenable, defaults to None
    :raise SystemExit: If there is an error with the request
    :return: Detailed list of Tenable scans
    :rtype: list
    """
    client = gen_client()
    detailed_scans = []
    for scan in track(scan_list, description="Fetching detailed scans..."):
        try:
            det = client.scans.details(id=scan["id"])
            detailed_scans.append(det)
        except RequestException as ex:  # This is the correct syntax
            raise SystemExit(ex) from ex

    return detailed_scans


@sc.command(name="save_queries")
@save_output_to()
@file_types([".json", ".csv", ".xlsx"])
def save_queries(save_output_to: Path, file_type: str):
    """Get a list of query definitions and save them as a .json, .csv or .xlsx file."""
    # get the queries from Tenable
    query_list = get_queries()

    # check if file path exists
    check_file_path(save_output_to)

    # set the file name
    file_name = f"tenable_queries_{get_current_datetime('%m%d%Y')}"

    # save the data as a .json file
    save_data_to(
        file=Path(f"{save_output_to}{os.sep}{file_name}{file_type}"),
        data=query_list,
    )


def get_queries() -> list:
    """
    List of query definitions

    :return: List of queries from Tenable
    :rtype: list
    """
    app = Application()
    tsc = gen_tsc(app.config)
    return tsc.queries.list()


@sc.command(name="query_vuln")
@click.option(
    "--query_id",
    type=click.INT,
    help="Tenable query ID to retrieve via API",
    prompt="Enter Tenable query ID",
    required=True,
)
@click.option(
    "--regscale_ssp_id",
    type=click.INT,
    help="The ID number from RegScale of the System Security Plan",
    prompt="Enter RegScale System Security Plan ID",
    required=True,
)
@click.option(
    "--create_issue_from_recommendation",
    type=click.BOOL,
    help="Create Issue in RegScale from Vulnerability in RegScale.",
    default=False,
    required=False,
)
# Add Prompt for RegScale SSP name
def query_vuln(query_id: int, regscale_ssp_id: int, create_issue_from_recommendation: bool):
    """Query Tenable vulnerabilities and sync assets to RegScale."""
    q_vuln(
        query_id=query_id,
        ssp_id=regscale_ssp_id,
        create_issue_from_recommendation=create_issue_from_recommendation,
    )


@io.command(name="sync_assets")
@click.option(
    "--regscale_ssp_id",
    type=click.INT,
    help="The ID number from RegScale of the System Security Plan",
    prompt="Enter RegScale System Security Plan ID",
    required=True,
)
# Add Prompt for RegScale SSP name
def query_assets(regscale_ssp_id: int):
    """Query Tenable Assets and sync to RegScale."""
    # Validate ssp

    if validate_regscale_security_plan(parent_id=regscale_ssp_id):
        fetch_assets(
            ssp_id=regscale_ssp_id,
        )
    else:
        raise ValueError("SSP ID does not exist in RegScale!")


def validate_regscale_security_plan(parent_id: int) -> bool:
    """
    Validate RegScale Security Plan exists

    :param int parent_id: The ID number from RegScale of the System Security Plan
    :return: If API call was successful
    :rtype: bool
    """
    app = check_license()
    config = app.config
    headers = {
        "Authorization": config["token"],
    }
    url = urljoin(config["domain"], f"/api/securityplans/{parent_id}")
    response = requests.get(url, headers=headers)
    return response.ok


@io.command(name="sync_vulns")
@click.option(
    "--regscale_ssp_id",
    type=click.INT,
    help="The ID number from RegScale of the System Security Plan",
    prompt="Enter RegScale System Security Plan ID",
    required=True,
)
@click.option(
    "--create_issue_from_recommendation",
    type=click.BOOL,
    help="Create RegScale issue from Tenable recommendation.",
    required=False,
)
# Add Prompt for RegScale SSP name
def query_vulns(regscale_ssp_id: int, create_issue_from_recommendation: bool = False):
    """Query Tenable Vulnerabilties and sync to RegScale."""
    # Validate ssp
    if validate_regscale_object(parent_id=regscale_ssp_id, parent_module="securityplans"):
        sync_vulns(
            regscale_ssp_id=regscale_ssp_id,
        )
        if create_issue_from_recommendation:
            create_regscale_issue_from_vuln(regscale_ssp_id=regscale_ssp_id)
    else:
        raise ValueError("SSP ID does not exist in RegScale!")


@io.command(name="list_jobs")
@click.option(
    "--job_type",
    default="vulns",
    type=click.Choice(["vulns", "assets"]),
    show_default=True,
    help="Tenable job type.",
    required=False,
)
@click.option(
    "--last",
    type=click.INT,
    default=100,
    show_default=True,
    help="Filter the last n jobs.",
    required=False,
)
@click.option(
    "--job_status",
    type=click.Choice(["processing", "finished", "cancelled"]),
    help="Filter by status.",
    required=False,
)
def list_jobs(job_type: str, last: int, job_status: str):
    """Retrieve a list of jobs from Tenable.io."""
    app = Application()
    config = app.config
    client = gen_tio(config)
    if job_status:
        jobs = [job for job in client.exports.jobs(job_type) if job["status"] == str(job_status).upper()]
    else:
        jobs = client.exports.jobs(job_type)
    jobs = sorted(jobs, key=lambda k: (k["created"]), reverse=False)
    # filter the last N jobs
    for job in jobs[len(jobs) - last :]:
        console.print(
            f"UUID: {job['uuid']}, STATUS: {job['status']}, CREATED: {epoch_to_datetime(job['created'], epoch_type='milliseconds')}"
        )


@io.command(name="cancel_job")
@click.option(
    "--uuid",
    type=click.STRING,
    help="Tenable job UUID.",
    prompt="Enter the UUID of the job to cancel.",
    required=True,
)
@click.option(
    "--job_type",
    default="vulns",
    type=click.Choice(["vulns", "assets"]),
    show_default=True,
    help="Tenable job type.",
    required=False,
)
def cancel_job(uuid: str, job_type: str):
    """Cancel a Tenable IO job."""
    app = Application()
    config = app.config
    client = gen_tio(config)
    client.exports.cancel(job_type, export_uuid=uuid)


@sc.command(name="trend_vuln")
@click.option(
    "-p",
    "--plugins",
    multiple=True,
    help="Enter one or more pluginID's separated by a space to see a trend-line. (by report date)",
    prompt="Enter one or more pluginID's",
    required=True,
)
@click.option(
    "-d",
    "--dnsname",
    multiple=False,
    type=click.STRING,
    help="Enter DNS name of asset to trend.",
    prompt="Enter DNS name of asset",
    required=True,
)
def trend_vuln(plugins: list, dnsname: str):
    """Trend vulnerabilities from vulnerability scans."""
    plugins = list(plugins)
    logger.info(plugins)
    trend_vulnerabilities(filter=plugins, dns=dnsname)


def process_vulnerabilities(counts: collections.Counter, reg_assets: list, ssp_id: int, tenable_vulns: list) -> list:
    """
    Process Tenable vulnerabilities

    :param collections.Counter counts: Dictionary of counts of each vulnerability
    :param list reg_assets: List of RegScale assets
    :param int ssp_id: RegScale System Security Plan ID
    :param list tenable_vulns: List of Tenable vulnerabilities
    :return: List of assets to update
    :rtype: list
    """
    update_assets = []
    for vuln in set(tenable_vulns):
        update_assets = process_vuln(counts, reg_assets, ssp_id, vuln)
    return update_assets


def create_and_update_assets(
    assets_to_be_inserted: list, insert_assets: list, update_assets: list, ssp_id: int
) -> Tuple[list, list]:
    """
    Create and update assets in RegScale

    :param list assets_to_be_inserted: List of assets to be inserted
    :param list insert_assets: List of assets to insert
    :param list update_assets: List of assets to update
    :param int ssp_id: RegScale System Security Plan ID
    :return: List of assets to insert, list of assets to update
    :rtype: Tuple[list, list]
    """
    app = Application()
    if assets_to_be_inserted:
        for t_asset in assets_to_be_inserted:
            r_asset = Asset(
                name=t_asset.dnsName,
                otherTrackingNumber=t_asset.pluginID,
                parentId=ssp_id,
                parentModule="securityplans",
                ipAddress=t_asset.ip,
                macAddress=t_asset.macAddress,
                assetOwnerId=app.config["userId"],
                status=get_status(t_asset),
                assetType="Other",
                assetCategory="Hardware",
                operatingSystem=determine_os(t_asset.operatingSystem),
            )
            if r_asset.macAddress.upper() not in {v["macAddress"].upper() for v in insert_assets}:
                insert_assets.append(r_asset)
    if insert_assets:
        res = Asset.batch_create(insert_assets)
        if res:
            logger.info("Successfully created %s new asset(s) in RegScale.", len(res))
    else:
        logger.info("No new assets need to be created in RegScale.")
    if update_assets:
        update_res = Asset.batch_update(update_assets)
        if update_res:
            logger.info("Successfully updated %s asset(s) in RegScale.", len(update_res))
    else:
        logger.info("No assets need to be updated in RegScale.")
    return insert_assets, update_assets


def q_vuln(query_id: int, ssp_id: int, create_issue_from_recommendation: bool) -> list:
    """
    Query Tenable vulnerabilities

    :param int query_id: Tenable query ID
    :param int ssp_id: RegScale System Security Plan ID
    :param bool create_issue_from_recommendation: Whether to create an issue in RegScale
    :return: List of queries from Tenable
    :rtype: list
    """
    check_license()
    # At SSP level, provide a list of vulnerabilities and the counts of each
    # Normalize the data based on mac address
    reg_assets = Asset.get_all_by_parent(parent_id=ssp_id, parent_module="securityplans")

    tenable_data = fetch_vulns(query_id=query_id, regscale_ssp_id=ssp_id)
    tenable_vulns = tenable_data[0]
    tenable_df = tenable_data[1]

    assets_to_be_inserted = list(
        {
            dat
            for dat in tenable_vulns
            if dat.macAddress not in {asset.macAddress for asset in inner_join(reg_assets, tenable_vulns)}
        }
    )
    counts = collections.Counter(s.pluginName for s in tenable_vulns)
    insert_assets = []
    update_assets = process_vulnerabilities(counts, reg_assets, ssp_id, tenable_vulns)

    insert_assets, update_assets = create_and_update_assets(assets_to_be_inserted, insert_assets, update_assets, ssp_id)

    if update_assets:
        update_res = Asset.batch_update(update_assets)
        if update_res:
            logger.info("Successfully updated %s asset(s) in RegScale.", len(update_res))
    else:
        logger.info("No assets need to be updated in RegScale.")

    if create_issue_from_recommendation and not tenable_df.empty:
        today = get_current_datetime(dt_format="%Y-%m-%d")
        create_regscale_issue_from_vuln(
            regscale_ssp_id=ssp_id,
            df=tenable_df[tenable_df["report_date"] == today],
        )
    else:
        logger.info("No issues need to be created or updated in RegScale.")
    return update_assets


def process_vuln(counts: collections.Counter, reg_assets: list, ssp_id: int, vuln: TenableAsset) -> list:
    """
    Process Tenable vulnerability data

    :param collections.Counter counts: Dictionary of counts of each vulnerability
    :param list reg_assets: List of RegScale assets
    :param int ssp_id: RegScale System Security Plan ID
    :param TenableAsset vuln: Tenable vulnerability object
    :return: List of assets to update
    :rtype: list
    """
    update_assets = []
    vuln.count = dict(counts)[vuln.pluginName]
    lookup_assets = lookup_asset(reg_assets, vuln.macAddress, vuln.dnsName)
    # Update parent id to SSP on insert
    if len(lookup_assets) > 0:
        for asset in set(lookup_assets):
            # Do update
            # asset = reg_asset[0]
            asset.parentId = ssp_id
            asset.parentModule = "securityplans"
            asset.macAddress = vuln.macAddress.upper()
            asset.osVersion = vuln.operatingSystem
            asset.purchaseDate = "01-01-1970"
            asset.endOfLifeDate = "01-01-1970"
            if asset.ipAddress is None:
                asset.ipAddress = vuln.ipAddress
            asset.operatingSystem = determine_os(asset.operatingSystem)
            try:
                assert asset.id
                # avoid duplication
                if asset not in update_assets:
                    update_assets.append(asset)
            except AssertionError as aex:
                logger.error("Asset does not have an id, unable to update!\n%s", aex)
    return update_assets


def determine_os(os_string: str) -> str:
    """
    Determine RegScale friendly OS name

    :param str os_string: String of the asset's OS
    :return: RegScale acceptable OS
    :rtype: str
    """
    linux_words = ["linux", "ubuntu", "hat", "centos", "rocky", "alma", "alpine"]
    if re.compile("|".join(linux_words), re.IGNORECASE).search(os_string):
        return "Linux"
    elif (os_string.lower()).startswith("windows"):
        return "Windows Server" if "server" in os_string else "Windows Desktop"
    else:
        return "Other"


def get_status(asset: TenableAsset) -> str:
    """
    Convert Tenable asset status to RegScale asset status

    :param TenableAsset asset: Asset object from Tenable
    :return: RegScale status
    :rtype: str
    """
    if asset.family.type == "active":
        return "Active (On Network)"
    return "Off-Network"  # Probably not correct


def lookup_asset(asset_list: list, mac_address: str, dns_name: str = None) -> list:
    """
    Lookup asset in Tenable and return the data from Tenable

    :param list asset_list: List of assets to lookup in Tenable
    :param str mac_address: Mac address of asset
    :param str dns_name: DNS Name of the asset, defaults to None
    :return: List of assets that fit the provided filters
    :rtype: list
    """
    results = []
    if validate_mac_address(mac_address):
        if dns_name:
            results = [
                Asset(**asset)
                for asset in asset_list
                if "macAddress" in asset
                and asset["macAddress"] == mac_address
                and asset["name"] == dns_name
                and "macAddress" in asset
                and "name" in asset
            ]
        else:
            results = [Asset(**asset) for asset in asset_list if asset["macAddress"] == mac_address]
    # Return unique list
    return list(set(results))


def trend_vulnerabilities(
    filter: list,
    dns: str,
    filter_type: str = "pluginID",
    filename: str = "vulnerabilities.pkl",
) -> None:
    """
    Trend vulnerabilities data to the console

    :param list filter: Data to use for trend graph
    :param str dns: DNS to filter data
    :param str filter_type: Type of filter to apply to data, defaults to "pluginID"
    :param str filename: Name of the file to save as, defaults to "vulnerabilities.pkl"
    :rtype: None
    """
    import pandas as pd  # Optimize import performance

    if not filter:
        return
    dataframe = pd.read_pickle(filename)
    dataframe = dataframe[dataframe[filter_type].isin(filter)]
    dataframe = dataframe[dataframe["dnsName"] == dns]
    unique_cols = ["pluginID", "dnsName", "severity", "report_date"]
    dataframe = dataframe[unique_cols]
    dataframe = dataframe.drop_duplicates(subset=unique_cols)
    if len(dataframe) == 0:
        error_and_exit("No Rows in Dataframe.")

    dataframe.loc[dataframe["severity"] == "Info", "severity_code"] = 0
    dataframe.loc[dataframe["severity"] == "Low", "severity_code"] = 1
    dataframe.loc[dataframe["severity"] == "Medium", "severity_code"] = 2
    dataframe.loc[dataframe["severity"] == "High", "severity_code"] = 3
    dataframe.loc[dataframe["severity"] == "Critical", "severity_code"] = 4
    # Deal with linux wayland sessions
    if "XDG_SESSION_TYPE" in os.environ and os.getenv("XDG_SESSION_TYPE") == "wayland":
        os.environ["QT_QPA_PLATFORM"] = "wayland"
    # plotting graph
    for filt in filter:
        plt.plot(dataframe["report_date"], dataframe["severity_code"], label=filt)
    logger.info("Plotting %s rows of data.\n", len(dataframe))
    logger.info(dataframe.head())
    plt.legend()
    plt.show(block=True)


def get_existing_regscale_issues(app: Application, regscale_ssp_id: int) -> List[dict]:
    """
    Gets existing RegScale issues for a System Security Plan

    :param Application app: Application object
    :param int regscale_ssp_id: RegScale System Security Plan ID
    :return: List of issues from RegScale
    :rtype: List[dict]
    """
    api = Api()
    existing_issues_req = api.get(app.config["domain"] + f"/api/issues/getAllByParent/{regscale_ssp_id}/securityplans")
    if existing_issues_req.status_code == 200:
        return existing_issues_req.json()
    else:
        return []


def create_issue_from_vuln(app: Application, row: "pd.Series", default_due_delta: int) -> "Issue":
    """
    Creates an Issue object from a Tenable vulnerability

    :param Application app: Application object
    :param pd.Series row: Row of data from Tenable
    :param int default_due_delta: Default due delta
    :return: Issue object
    :rtype: Issue
    """

    default_status = app.config["issues"]["tenable"]["status"]
    fmt = "%Y-%m-%d %H:%M:%S"
    plugin_id = row["pluginID"]
    port = row["port"]
    protocol = row["protocol"]
    due_date = datetime.strptime(row["last_scan"], fmt) + timedelta(days=default_due_delta)
    if due_date < datetime.now():
        due_date = datetime.now() + timedelta(days=default_due_delta)
    if "synopsis" in row:
        title = row["synopsis"]
    return Issue(
        title=title or row["pluginName"],
        description=row["description"] or row["pluginName"] + f"<br>Port: {port}<br>Protocol: {protocol}",
        issueOwnerId=app.config["userId"],
        status=default_status,
        severityLevel=Issue.assign_severity(row["severity"]),
        dueDate=due_date.strftime(fmt),
        identification="Vulnerability Assessment",
        parentId=row["regscale_ssp_id"],
        parentModule="securityplans",
        pluginId=plugin_id,
        vendorActions=row["solution"],
        assetIdentifier=f'DNS: {row["dnsName"]} - IP: {row["ip"]}',
    )


def create_issue_from_row(app: Application, row: "pd.Series", default_due_delta: int) -> "Issue":
    """
    Creates an Issue object from a Tenable vulnerability

    :param Application app: Application object
    :param pd.Series row: Row of data from Tenable
    :param int default_due_delta: Default due delta
    :return: Issue object
    :rtype: Issue
    """
    if row["severity"] != "Info":
        issue = create_issue_from_vuln(app, row, default_due_delta)
        if isinstance(issue, Issue):
            return issue
    return None


def prepare_issues_for_sync(
    app: Application, df: "pd.DataFrame", regscale_ssp_id: int
) -> Tuple[List["Issue"], List["Issue"]]:
    """
    Prepares Tenable vulnerabilities for synchronization as RegScale issues

    :param Application app: Application object
    :param pd.DataFrame df: Dataframe of Tenable data
    :param int regscale_ssp_id: RegScale System Security Plan ID
    :return: List of issues to insert, list of issues to update
    :rtype: Tuple[List[Issue], List[Issue]]
    """

    default_due_delta = app.config["issues"]["tenable"]["moderate"]
    existing_issues = Issue.get_all_by_parent(parent_id=regscale_ssp_id, parent_module="securityplans")
    sc_issues = []
    new_issues = set()
    update_issues = set()
    for index, row in df.iterrows():
        issue = create_issue_from_row(app, row, default_due_delta)
        if isinstance(issue, Issue):
            sc_issues.append(issue)
    # Generate list of completely new issues, and merge with existing issues if they have the same title
    # group issues by title
    grouped_issues = {k: list(g) for k, g in groupby(sc_issues, key=lambda x: x.title)}
    for title in grouped_issues:
        reg_key = 0
        regs = [iss for iss in existing_issues if iss.title == title and iss.id]
        if regs:
            reg_key = regs[0].id
        issues = set(grouped_issues[title])
        for issue in issues:
            asset_ident = combine_strings({iss.assetIdentifier for iss in issues})
            if reg_key and issue.title not in {iss.title for iss in update_issues}:
                issue.id = reg_key
                issue.assetIdentifier = asset_ident
                update_issues.add(issue)
            elif not reg_key:
                issue.assetIdentifier = asset_ident
                new_issues.add(issue)

    return list(new_issues), list(update_issues)


def combine_strings(set_of_strings: Set[str]) -> str:
    """
    Combines a set of strings into a single string

    :param Set[str] set_of_strings: Set of strings
    :rtype: str
    :return: Combined string
    """
    return "<br>".join(set_of_strings)


def sync_issues_to_regscale(new_issues: List["Issue"], update_issues: List["Issue"]) -> None:
    """
    Synchronizes issues to RegScale

    :param List[Issue] new_issues: New issues
    :param List[Issue] update_issues: Updated issues
    :rtype: None
    """
    logger = create_logger()

    if new_issues:
        logger.info(f"Creating {len(new_issues)} new issue(s) in RegScale...")
        Issue.batch_create(new_issues)
        logger.info("Finished creating issue(s) in RegScale.")
    else:
        logger.info("No new issues to create.")

    if update_issues:
        logger.info(f"Updating {len(update_issues)} existing issue(s) in RegScale...")
        Issue.batch_update(update_issues)
        logger.info("Finished updating issue(s) in RegScale.")
    else:
        logger.info("No issues to update.")


def create_regscale_issue_from_vuln(regscale_ssp_id: int, df: Optional["pd.DataFrame"] = None) -> None:
    """
    Sync Tenable Vulnerabilities to RegScale issues

    :param int regscale_ssp_id: RegScale System Security Plan ID
    :param Optional["pd.DataFrame"] df: Pandas dataframe of Tenable data
    :rtype: None
    """
    if not df:
        import pandas as pd

        df = pd.DataFrame()
    app = Application()
    new_issues, update_issues = prepare_issues_for_sync(app, df, regscale_ssp_id)
    sync_issues_to_regscale(new_issues, update_issues)


def log_vulnerabilities(data: list, query_id: int, regscale_ssp_id: int) -> "pd.DataFrame":
    """
    Logs Vulnerabilities to a panda's dataframe

    :param list data: Raw data of Tenable vulnerabilities
    :param int query_id: Query ID for Tenable query
    :param int regscale_ssp_id: RegScale System Security Plan ID
    :return: Cleaned up data of Tenable vulnerabilities
    :rtype: pd.DataFrame
    """
    import pandas as pd  # Optimize import performance

    warnings.filterwarnings("ignore", category=FutureWarning)
    try:
        dataframe = pd.DataFrame([x.dict() for x in data])
        if dataframe.empty:
            return dataframe
        dataframe["query_id"] = query_id
        dataframe["regscale_ssp_id"] = regscale_ssp_id
        dataframe["first_scan"] = dataframe["firstSeen"].apply(epoch_to_datetime)
        dataframe["last_scan"] = dataframe["lastSeen"].apply(epoch_to_datetime)
        dataframe["severity"] = [d.get("name") for d in dataframe["severity"]]
        dataframe["family"] = [d.get("name") for d in dataframe["family"]]
        dataframe["repository"] = [d.get("name") for d in dataframe["repository"]]
        dataframe["report_date"] = get_current_datetime(dt_format="%Y-%m-%d")
        filename = "vulnerabilities.pkl"

        dataframe.drop_duplicates()
        if not Path(filename).exists():
            logger.info("Saving vulnerability data to %s.", filename)
        else:
            logger.info("Updating vulnerabilities.pkl with the latest data from Tenable.")
            old_df = pd.read_pickle(filename)
            old_df = old_df[old_df["report_date"] != get_current_datetime(dt_format="%Y-%m-%d")]
            try:
                dataframe = pd.concat([old_df, dataframe]).drop_duplicates()
            except ValueError as vex:
                logger.error("Pandas ValueError:%s.", vex)
        dataframe.to_pickle(filename)
        severity_arr = dataframe.groupby(["severity", "repository"]).size().to_frame()
        console.rule("[bold red]Vulnerability Overview")
        console.print(severity_arr)
        return dataframe

    except pd.errors.DataError as dex:
        logger.error(dex)


def fetch_assets(ssp_id: int) -> list[TenableIOAsset]:
    """
    Fetch assets from Tenable IO and sync to RegScale

    :param int ssp_id: RegScale System Security Plan ID
    :return: List of Tenable assets
    :rtype: list[TenableIOAsset]
    """
    tenable_last_updated: int = 0
    app = Application()
    config = app.config
    client = gen_tio(config=config)
    assets: List[TenableIOAsset] = []
    logger.info("Fetching existing assets from RegScale...")

    existing_assets: List[Asset] = Asset.get_all_by_parent(parent_id=ssp_id, parent_module="securityplans")

    logger.info("Found %i existing asset(s) in RegScale.", len(existing_assets))

    filtered_assets = [asset for asset in existing_assets if asset.tenableId and asset.dateLastUpdated]
    # Get last epoch updated from RegScale, limit to Tenable assets
    if filtered_assets:
        tenable_last_updated = max([regscale_string_to_epoch(asset.dateLastUpdated) for asset in filtered_assets])
    export = client.exports.assets(updated_at=tenable_last_updated)
    logger.info("Saving chunked asset files from Tenable IO for processing...")
    temp_loc = Path(tempfile.gettempdir()) / "tenable_io" / str(uuid.uuid4())  # random folder name
    # show process status
    box_len = 0
    status = client.exports.status(export_type=export.type, export_uuid=export.uuid)
    with create_progress_object(indeterminate=True) as job_progress:
        job_progress.add_task("Fetching Chunked Tenable IO data...", start=False, total=None)
        while status["status"] == "PROCESSING":
            box_len = len(status["chunks_available"])
            time.sleep(0.5)
            status = client.exports.status(export_type=export.type, export_uuid=export.uuid)
    # Process chunks of data
    with create_progress_object(indeterminate=True) as saving_progress:
        saving_task = saving_progress.add_task(
            "Saving Tenable IO data to disk...",
            total=box_len,
        )
        export.run_threaded(
            func=write_io_chunk,
            kwargs={"data_dir": temp_loc},
            num_threads=10,
        )
        saving_progress.update(saving_task, advance=1)
    process_to_regscale(data_dir=temp_loc, ssp_id=ssp_id, existing_assets=existing_assets)
    return assets


def fetch_vulns(query_id: int = 0, regscale_ssp_id: int = 0) -> Tuple[list, "pd.DataFrame"]:
    """
    Fetch vulnerabilities from Tenable by query ID

    :param int query_id: Tenable query ID, defaults to 0
    :param int regscale_ssp_id: RegScale System Security Plan ID, defaults to 0
    :return: Tuple[list of vulnerabilities from Tenable, Tenable vulnerabilities as a panda's dataframe]
    :rtype: Tuple[list, pd.DataFrame]
    """
    client = gen_client()
    data: List[TenableAsset] = []
    if query_id:
        description = f"Fetching Vulnerabilities for Tenable query id: {query_id}."
        if client._env_base == "TSC":
            vulns = client.analysis.vulns(query_id=query_id)
            data.extend(
                TenableAsset.from_dict(vuln) for vuln in track(vulns, description=description, show_speed=False)
            )
            logger.info("Found %i vulnerabilities.", len(data))
    dataframe = log_vulnerabilities(data, query_id=query_id, regscale_ssp_id=regscale_ssp_id)
    return data, dataframe


def get_last_pull_epoch(regscale_ssp_id: int) -> int:
    """
    Gather last pull epoch from RegScale Security Plan

    :param int regscale_ssp_id: RegScale System Security Plan ID
    :return: Last pull epoch
    :rtype: int

    """
    fmt: str = "%Y-%m-%d"
    two_months_ago: datetime = datetime.now() - timedelta(weeks=8)
    two_weeks_ago: datetime = datetime.now() - timedelta(weeks=2)
    last_pull: int = round(two_weeks_ago.timestamp())  # default the last pull date to two weeks
    # Limit the query with a filter_date to avoid taxing the database in the case of a large number of scans
    if res := ScanHistory.get_by_parent_recursive(
        parent_id=regscale_ssp_id, parent_module="securityplans", filter_date=two_months_ago.strftime(fmt)
    ):
        # order by ScanDate desc
        fmt = "%Y-%m-%dT%H:%M:%S"
        res = sorted(res, key=lambda x: datetime.strptime(x.scanDate, fmt), reverse=True)
        # Convert to timestampe
        last_pull = round((datetime.strptime(res[0].scanDate, fmt)).timestamp())
    return last_pull


def sync_vulns(regscale_ssp_id: int = 0) -> None:
    """
    Sync Vulnerabilities

    :param int regscale_ssp_id: _description_, defaults to 0
    :rtype: None
    """
    from regscale.models.integration_models.nessus import NessusReport

    app = Application()
    config = app.config
    client = gen_tio(config=config)
    minimum_severity = validate_nessus_severity(config["tenableMinimumSeverityFilter"])
    cvss2_min = get_min_cvss_score(minimum_severity)
    latest_scan: int = get_last_pull_epoch(regscale_ssp_id)
    vulns = client.exports.vulns(vpr_score={"gte": cvss2_min}, since=latest_scan, state=["OPEN", "REOPENED"])
    # Chunks are typically 200MB a piece, so we can use this to determine a rough estimate of the space needed
    console.print(
        f"The space available on the temp share ({tempfile.gettempdir()}) for caching Tenable data is "
        f"{str(round(determine_available_space() / (1024 ** 3), 1))} GB.",
    )
    if latest_scan:
        app.logger.info(
            "Building a list of Nessus Reports with a severity >= %s and updated since %s.",
            minimum_severity,
            datetime.fromtimestamp(latest_scan),
        )
    else:
        app.logger.info("Building a list of Nessus Reports with a severity >= %s.", minimum_severity)
    NessusReport.sync_to_regscale(vulns=vulns, parent_id=regscale_ssp_id, client=client, config=config)
    logger.info("Done!")


@sc.command(name="list_tags")
def sc_tags():
    """List tags from Tenable"""
    list_tags()


def list_tags() -> None:
    """
    Query a list of tags on the server and print to console

    :rtype: None
    """
    tag_list = get_tags()
    pprint(tag_list)


def get_tags() -> list:
    """
    List of Tenable query definitions

    :return: List of unique tags for Tenable queries
    :rtype: list
    """
    client = gen_client()
    logger.debug(client._env_base)
    if client._env_base == "TSC":
        return client.queries.tags()
    return list(client.tags.list())


def gen_client() -> Union[TenableIO, TenableSC]:
    """
    Return the appropriate Tenable client based on the URL

    :return: Client type
    :rtype: Union[TenableIO, TenableSC]
    """
    app = Application()
    config = app.config
    if "cloud.tenable.com" in config["tenableUrl"]:
        return gen_tio(config)
    return gen_tsc(config)


def gen_tsc(config: dict) -> TenableSC:
    """
    Generate Tenable Object

    :param dict config: Configuration dictionary
    :return: Tenable client
    :rtype: TenableSC
    """
    if not config:
        app = Application()
        config = app.config
    return TenableSC(
        url=config["tenableUrl"],
        access_key=config["tenableAccessKey"],
        secret_key=config["tenableSecretKey"],
        vendor=REGSCALE_INC,
        product=REGSCALE_CLI,
        build=__version__,
    )


def gen_tio(config: dict) -> TenableIO:
    """
    Generate Tenable Object

    :param dict config: Configuration dictionary
    :return: Tenable client
    :rtype: TenableIO
    """
    return TenableIO(
        url=config["tenableUrl"],
        access_key=config["tenableAccessKey"],
        secret_key=config["tenableSecretKey"],
        vendor=REGSCALE_INC,
        product=REGSCALE_CLI,
        build=__version__,
    )


def inner_join(reg_list: list[Asset], tenable_list: list) -> list:
    """
    Function to inner join two lists on the macAddress field and returns the assets existing in RegScale and Tenable

    :param list[Asset] reg_list: List of RegScale assets
    :param list tenable_list: List of Tenable assets
    :return: Returns list of assets that exists in RegScale and Tenable using the mac address
    :rtype: list
    """
    set1 = {lst["macAddress"].lower() for lst in reg_list if lst.macAddress}
    data = []
    try:
        data = [list_ten for list_ten in tenable_list if list_ten.macAddress.lower() in set1]
    except KeyError as ex:
        logger.error(ex)
    return data


def get_control_implementations(parent_id: int) -> list:
    """
    Gets all the control implementations

    :param parent_id: parent control id
    :type parent_id: int
    :return: list of control implementations
    :rtype: list
    """
    app = Application()
    api = Api()
    url = urljoin(app.config.get("domain"), f"/api/controlImplementation/getAllByPlan/{parent_id}")
    response = api.get(url)
    if response.ok:
        return response.json()
    else:
        response.raise_for_status()
    return []


def get_controls(catalog_id: int) -> List[Dict]:
    """
    Gets all the controls

    :param int catalog_id: catalog id
    :return: list of controls
    :rtype: List[Dict]
    """
    app = Application()
    api = Api()
    url = urljoin(app.config.get("domain"), f"/api/SecurityControls/getList/{catalog_id}")
    response = api.get(url)
    if response.ok:
        return response.json()
    else:
        response.raise_for_status()
    return []


def create_control_implementations(
    controls: list,
    parent_id: int,
    parent_module: str,
    existing_implementation_dict: Dict,
    passing_controls: Dict,
    failing_controls: Dict,
) -> List[Dict]:
    """
    Creates a list of control implementations

    :param list controls: list of controls
    :param int parent_id: parent control id
    :param str parent_module: parent module
    :param Dict existing_implementation_dict: Dictionary of existing control implementations
    :param Dict passing_controls: Dictionary of passing controls
    :param Dict failing_controls: Dictionary of failing controls
    :return: list of control implementations
    :rtype: List[Dict]
    """
    app = Application()
    api = Api()
    user_id = app.config.get("userId")
    domain = app.config.get("domain")
    control_implementations = []
    to_create = []
    to_update = []
    for control in controls:
        lower_case_control_id = control["controlId"].lower()
        status = check_implementation(
            passing_controls=passing_controls,
            failing_controls=failing_controls,
            control_id=lower_case_control_id,
        )
        if control["controlId"] not in existing_implementation_dict.keys():
            cim = ControlImplementation(
                controlOwnerId=user_id,
                dateLastAssessed=get_current_datetime(),
                status=status,
                controlID=control["id"],
                parentId=parent_id,
                parentModule=parent_module,
                createdById=user_id,
                dateCreated=get_current_datetime(),
                lastUpdatedById=user_id,
                dateLastUpdated=get_current_datetime(),
            ).dict()
            cim["controlSource"] = "Baseline"
            to_create.append(cim)

        else:
            # update existing control implementation data
            existing_imp = existing_implementation_dict.get(control["controlId"])
            existing_imp["status"] = status
            existing_imp["dateLastAssessed"] = get_current_datetime()
            existing_imp["lastUpdatedById"] = user_id
            existing_imp["dateLastUpdated"] = get_current_datetime()
            del existing_imp["createdBy"]
            del existing_imp["systemRole"]
            del existing_imp["controlOwner"]
            del existing_imp["lastUpdatedBy"]
            to_update.append(existing_imp)

    if len(to_create) > 0:
        ci_url = urljoin(domain, "/api/controlImplementation/batchCreate")
        resp = api.post(url=ci_url, json=to_create)
        if resp.ok:
            control_implementations.extend(resp.json())
            logger.info(f"Created {len(to_create)} Control Implementation(s), Successfully!")
        else:
            resp.raise_for_status()
    if len(to_update) > 0:
        ci_url = urljoin(domain, "/api/controlImplementation/batchUpdate")
        resp = api.post(url=ci_url, json=to_update)
        if resp.ok:
            control_implementations.extend(resp.json())
            logger.info(f"Updated {len(to_update)} Control Implementation(s), Successfully!")
        else:
            resp.raise_for_status()
    return control_implementations


def check_implementation(passing_controls: Dict, failing_controls: Dict, control_id: str) -> str:
    """
    Checks the status of a control implementation

    :param Dict passing_controls: Dictionary of passing controls
    :param Dict failing_controls: Dictionary of failing controls
    :param str control_id: control id
    :return: status of control implementation
    :rtype: str
    """
    if control_id in passing_controls.keys():
        return FULLY_IMPLEMENTED
    elif control_id in failing_controls.keys():
        return IN_REMEDIATION
    else:
        return NOT_IMPLEMENTED


def get_existing_control_implementations(parent_id: int) -> Dict:
    """
    fetch existing control implementations

    :param int parent_id: parent control id
    :return: Dictionary of existing control implementations
    :rtype: Dict
    """
    app = Application()
    api = Api()
    domain = app.config.get("domain")
    existing_implementation_dict = {}
    get_url = urljoin(domain, f"/api/controlImplementation/getAllByPlan/{parent_id}")
    response = api.get(get_url)
    if response.ok:
        existing_control_implementations_json = response.json()
        for cim in existing_control_implementations_json:
            existing_implementation_dict[cim["controlName"]] = cim
        logger.info(f"Found {len(existing_implementation_dict)} existing control implementations")
    elif response.status_code == 404:
        logger.info(f"No existing control implementations found for {parent_id}")
    else:
        logger.warn(f"Unable to get existing control implementations. {response.text}")

    return existing_implementation_dict


def get_matched_controls(tenable_controls: List[Dict], catalog_controls: List[Dict]) -> List[Dict]:
    """
    Get controls that match between Tenable and the catalog

    :param List[Dict] tenable_controls: List of controls from Tenable
    :param List[Dict] catalog_controls: List of controls from the catalog
    :return: List of matched controls
    :rtype: List[Dict]
    """
    matched_controls = []
    for control in tenable_controls:
        formatted_control = convert_control_id(control)
        logger.info(formatted_control)
        for catalog_control in catalog_controls:
            if catalog_control["controlId"].lower() == formatted_control.lower():
                logger.info(f"Catalog Control {formatted_control} matched")
                matched_controls.append(catalog_control)
                break
    return matched_controls


def get_assessment_status_from_implementation_status(status: str) -> str:
    """
    Get the assessment status from the implementation status

    :param str status: Implementation status
    :return: Assessment status
    :rtype: str
    """
    if status == FULLY_IMPLEMENTED:
        return "Pass"
    elif status == IN_REMEDIATION:
        return "Fail"
    else:
        return "N/A"


def create_assessment_from_cim(cim: Dict, user_id: str, control: Dict, check: List[AssetCheck]) -> Dict:
    """
    Create an assessment from a control implementation

    :param Dict cim: Control Implementation
    :param str user_id: User ID
    :param Dict control: Control
    :param List[AssetCheck] check: Asset Check
    :return: Assessment
    :rtype: Dict
    """
    assessment_result = get_assessment_status_from_implementation_status(cim.get("status"))
    summary_dict = check[0].dict() if check else dict()
    summary_dict.pop("reference", None)
    title = summary_dict.get("check_name") if summary_dict else control.get("title")
    html_summary = format_dict_to_html(summary_dict)
    document_reviewed = check[0].audit_file if check else None
    check_name = check[0].check_name if check else None
    methodology = check[0].check_info if check else None
    summary_of_results = check[0].description if check else None
    uuid = check[0].asset_uuid if check and check[0].asset_uuid is not None else None
    title_part = f"{title} - {uuid}" if uuid else f"{title}"
    uuid_title = f"{title_part} Automated Assessment test"
    return {
        "leadAssessorId": user_id,
        "title": uuid_title,
        "assessmentType": "Control Testing",
        "plannedStart": get_current_datetime(),
        "plannedFinish": get_current_datetime(),
        "status": "Complete",
        "assessmentResult": assessment_result if assessment_result else "N/A",
        "controlID": cim["id"],
        "actualFinish": get_current_datetime(),
        "assessmentReport": html_summary if html_summary else "Passed",
        "parentId": cim["id"],
        "parentModule": "controls",
        "assessmentPlan": check_name if check_name else None,
        "documentsReviewed": document_reviewed if document_reviewed else None,
        "methodology": methodology if methodology else None,
        "summaryOfResults": summary_of_results if summary_of_results else None,
    }


def get_control_assessments(control: Dict, assessments_to_create: List[Dict]) -> List[Dict]:
    """
    Get control assessments

    :param Dict control: Control
    :param List[Dict] assessments_to_create: List of assessments to create
    :return: List of control assessments
    :rtype: List[Dict]
    """
    return [
        assess
        for assess in assessments_to_create
        if assess["controlID"] == control["id"] and assess["status"] == "Complete"
    ]


def sort_assessments(control_assessments: List[Dict]) -> List[Dict]:
    """
    Sort assessments by actual finish date

    :param List[Dict] control_assessments: List of control assessments
    :return: Sorted assessments
    :rtype: List[Dict]
    """
    dt_format = "%Y-%m-%d %H:%M:%S"
    return sorted(
        control_assessments,
        key=lambda x: datetime.strptime(x["actualFinish"], dt_format),
        reverse=True,
    )


def update_control_object(control: Dict, sorted_assessments: List[Dict]) -> None:
    """
    Update control object

    :param Dict control: Control
    :param List[Dict] sorted_assessments: Sorted assessments
    :rtype: None
    """
    dt_format = "%Y-%m-%d %H:%M:%S"
    app = Application()
    control["dateLastAssessed"] = sorted_assessments[0]["actualFinish"]
    control["lastAssessmentResult"] = sorted_assessments[0]["assessmentResult"]
    if control.get("lastAssessmentResult"):
        control_obj = ControlImplementation(**control)
        if control_obj.lastAssessmentResult == "Fail" and control_obj.status != IN_REMEDIATION:
            control_obj.status = IN_REMEDIATION
            control_obj.plannedImplementationDate = (datetime.now() + timedelta(30)).strftime(dt_format)
            control_obj.stepsToImplement = "n/a"
        elif control_obj.status == IN_REMEDIATION:
            control_obj.plannedImplementationDate = (
                (datetime.now() + timedelta(30)).strftime(dt_format)
                if not control_obj.plannedImplementationDate
                else control_obj.plannedImplementationDate
            )
            control_obj.stepsToImplement = "n/a" if not control_obj.stepsToImplement else control_obj.stepsToImplement
        elif control_obj.lastAssessmentResult == "Pass" and control_obj.status != FULLY_IMPLEMENTED:
            control_obj.status = FULLY_IMPLEMENTED
        ControlImplementation.update(app=app, implementation=control_obj)


def update_control_implementations(control_implementations: List[Dict], assessments_to_create: List[Dict]) -> None:
    """
    Update control implementations with assessments

    :param List[Dict] control_implementations: List of control implementations
    :param List[Dict] assessments_to_create: List of assessments to create
    :rtype: None
    """
    for control in control_implementations:
        control_assessments = get_control_assessments(control, assessments_to_create)
        if sorted_assessments := sort_assessments(control_assessments):
            update_control_object(control, sorted_assessments)


def post_assessments_to_api(assessments_to_create: List[Dict]) -> None:
    """
    Post assessments to the API

    :param List[Dict] assessments_to_create: List of assessments to create
    :rtype: None
    """
    app = Application()
    api = Api()
    assessment_url = urljoin(app.config.get("domain", ""), "/api/assessments/batchCreate")
    assessment_response = api.post(url=assessment_url, json=assessments_to_create)
    if assessment_response.ok:
        logger.info(f"Created {len(assessment_response.json())} Assessments!")
    else:
        logger.debug(assessment_response.status_code)
        logger.error(f"Failed to insert Assessment.\n{assessment_response.text}")


def create_assessments(
    control_implementations: List[Dict],
    catalog_controls_dict: Dict,
    asset_checks: Dict[str, List[AssetCheck]],
) -> None:
    """
    Create assessments from control implementations

    :param List[Dict] control_implementations: List of control implementations
    :param Dict catalog_controls_dict: Dictionary of catalog controls
    :param Dict[str, List[AssetCheck]] asset_checks: Dictionary of asset checks
    :rtype: None
    """
    app = Application()
    user_id = app.config.get("userId", "")
    assessments_to_create = []
    for cim in control_implementations:
        control = catalog_controls_dict.get(cim["controlID"], {})
        check = asset_checks.get(control["controlId"].lower())
        assessment = create_assessment_from_cim(cim, user_id, control, check)
        assessments_to_create.append(assessment)
    update_control_implementations(control_implementations, assessments_to_create)
    post_assessments_to_api(assessments_to_create)


def get_future_datetime() -> str:
    """
    Get a future date and time of today + 30 days as a string

    :return: Today + 30 days as a string in the format "YYYY-MM-DDTHH:MM:SSZ"
    :rtype: str
    """
    # Get the current date and time
    current_datetime = datetime.now()

    # Add 30 days to the current date and time
    future_datetime = current_datetime + timedelta(days=30)
    # Convert the datetime object to a string in the format "YYYY-MM-DDTHH:MM:SSZ"
    future_datetime_str = future_datetime.strftime("%Y-%m-%dT%H:%M:%SZ")

    return future_datetime_str


def map_status(status: str) -> str:
    """
    Map the status from Tenable.io to the status in the catalog

    :param str status: The status from Tenable.io
    :return: The status in the catalog
    :rtype: str
    """
    if status == "FAILED":
        return "Fail"
    elif status == "PASSED":
        return "Pass"
    else:
        return "N/A"


def process_compliance_data(
    framework_data: Dict,
    catalog_id: int,
    ssp_id: int,
    framework: str,
    passing_controls: Dict,
    failing_controls: Dict,
) -> None:
    """
    Processes the compliance data from Tenable.io to create control implementations for controls in frameworks

    :param Dict framework_data: List of tenable.io controls per framework
    :param int catalog_id: The catalog id
    :param int ssp_id: The ssp id
    :param str framework: The framework name
    :param Dict passing_controls: Dictionary of passing controls
    :param Dict failing_controls: Dictionary of failing controls
    :rtype: None
    """
    if not framework_data:
        return
    framework_controls = framework_data.get("controls", {})
    asset_checks = framework_data.get("asset_checks", {})
    existing_implementation_dict = get_existing_control_implementations(ssp_id)
    catalog_controls = get_controls(catalog_id)
    matched_controls = []
    for tenable_framework, tenable_controls in framework_controls.items():
        logger.info(f"Found {len(tenable_controls)} controls that passed for framework: {tenable_framework}")
        # logger.info(f"tenable_controls: {tenable_controls[0]}") if len(tenable_controls) >0 else None
        if tenable_framework == framework:
            matched_controls = get_matched_controls(tenable_controls, catalog_controls)

    logger.info(f"Found {len(matched_controls)} controls that matched")

    control_implementations = create_control_implementations(
        controls=matched_controls,
        parent_id=ssp_id,
        parent_module="securityplans",
        existing_implementation_dict=existing_implementation_dict,
        passing_controls=passing_controls,
        failing_controls=failing_controls,
    )

    logger.info(f"SSP now has {len(control_implementations)} control implementations")
    catalog_controls_dict = {c["id"]: c for c in catalog_controls}
    create_assessments(control_implementations, catalog_controls_dict, asset_checks)


def convert_control_id(control_id: str) -> str:
    """
    Convert the control id to a format that can be used in Tenable.io

    :param str control_id: The control id to convert
    :return: The converted control id
    :rtype: str
    """
    # Convert to lowercase
    control_id = control_id.lower()

    # Check if there's a parenthesis and replace its content
    if "(" in control_id and ")" in control_id:
        inner_value = control_id.split("(")[1].split(")")[0]
        control_id = control_id.replace(f"({inner_value})", f".{inner_value}")

    return control_id


@io.command(name="sync_compliance_controls")
@click.option(
    "--regscale_ssp_id",
    type=click.INT,
    help="The ID number from RegScale of the System Security Plan",
    prompt="Enter RegScale System Security Plan ID",
    required=True,
)
@click.option(
    "--catalog_id",
    type=click.INT,
    help="The ID number from RegScale Catalog that the System Security Plan's controls belong to",
    prompt="Enter RegScale Catalog ID",
    required=True,
)
@click.option(
    "--framework",
    required=True,
    type=click.Choice(["800-53", "800-53r5", "CSF", "800-171"], case_sensitive=True),
    help="The framework to use. from Tenable.io frameworks MUST be the same RegScale Catalog of controls",
)
def sync_compliance_data(regscale_ssp_id: int, catalog_id: int, framework: str):
    """
    Sync the compliance data from Tenable.io to create control implementations for controls in frameworks.
    """
    _sync_compliance_data(ssp_id=regscale_ssp_id, catalog_id=catalog_id, framework=framework)


def _sync_compliance_data(ssp_id: int, catalog_id: int, framework: str) -> None:
    """
    Sync the compliance data from Tenable.io to create control implementations for controls in frameworks
    :param int ssp_id: The ID number from RegScale of the System Security Plan
    :param int catalog_id: The ID number from RegScale Catalog that the System Security Plan's controls belong to
    :param str framework: The framework to use. from Tenable.io frameworks MUST be the same RegScale Catalog of controls
    :rtype: None
    """
    logger.info("Note: This command only available for Tenable.io")
    logger.info("Note: This command Requires admin access.")
    app = Application()
    config = app.config
    # we specifically don't gen client here, so we only get the client for Tenable.io as its only supported there
    client = TenableIO(
        url="https://cloud.tenable.com",
        access_key=config["tenableAccessKey"],
        secret_key=config["tenableSecretKey"],
        vendor=REGSCALE_INC,
        product=REGSCALE_CLI,
        build=__version__,
    )

    dict_of_frameworks_and_asset_checks: Dict = dict()
    framework_controls: Dict[str, List[str]] = {}
    asset_checks: Dict[str, List[AssetCheck]] = {}
    passing_controls: Dict = dict()
    # partial_passing_controls: Dict = dict()
    failing_controls: Dict = dict()
    for findings in client.exports.compliance():
        asset_check = AssetCheck(**findings)
        for ref in asset_check.reference:
            if ref.framework not in framework_controls:
                framework_controls[ref.framework] = []
            if ref.control not in framework_controls[ref.framework]:  # Avoid duplicate controls
                framework_controls[ref.framework].append(ref.control)
                formatted_control_id = convert_control_id(ref.control)
                # sort controls by status
                add_control_to_status_dict(
                    control_id=formatted_control_id,
                    status=asset_check.status,
                    dict_obj=failing_controls,
                    desired_status="FAILED",
                )
                add_control_to_status_dict(
                    control_id=formatted_control_id,
                    status=asset_check.status,
                    dict_obj=passing_controls,
                    desired_status="PASSED",
                )
                remove_passing_controls_if_in_failed_status(passing=passing_controls, failing=failing_controls)
                if formatted_control_id not in asset_checks:
                    asset_checks[formatted_control_id] = [asset_check]
                else:
                    asset_checks[formatted_control_id].append(asset_check)
        dict_of_frameworks_and_asset_checks = {
            key: {"controls": framework_controls, "asset_checks": asset_checks} for key in framework_controls.keys()
        }
    logger.info(f"Found {len(dict_of_frameworks_and_asset_checks)} findings to process")
    framework_data = dict_of_frameworks_and_asset_checks.get(framework, None)
    process_compliance_data(
        framework_data=framework_data,
        catalog_id=catalog_id,
        ssp_id=ssp_id,
        framework=framework,
        passing_controls=passing_controls,
        failing_controls=failing_controls,
    )


def add_control_to_status_dict(control_id: str, status: str, dict_obj: Dict, desired_status: str) -> None:
    """
    Add a control to a status dictionary

    :param str control_id: The control id to add to the dictionary
    :param str status: The status of the control
    :param Dict dict_obj: The dictionary to add the control to
    :param str desired_status: The desired status of the control
    :rtype: None
    """
    friendly_control_id = control_id.lower()
    if status == desired_status and friendly_control_id not in dict_obj:
        dict_obj[friendly_control_id] = desired_status


def remove_passing_controls_if_in_failed_status(passing: Dict, failing: Dict) -> None:
    """
    Remove passing controls if they are in failed status

    :param Dict passing: Dictionary of passing controls
    :param Dict failing: Dictionary of failing controls
    :rtype: None
    """
    to_remove = []
    for k in passing.keys():
        if k in failing.keys():
            to_remove.append(k)

    for k in to_remove:
        del passing[k]


def write_io_chunk(
    data: List[dict],
    data_dir: Path,
    export_uuid: str,
    export_type: str,
    export_chunk_id: int,
) -> None:
    """
    Write a chunk of data to a file, this function is formatted for use with PyTenable and Tenable IO

    :param List[dict] data: Data to write to a file
    :param Path data_dir: Directory to write the file to
    :param str export_uuid: UUID of the export (Tenable IO)
    :param str export_type: Type of export (Tenable IO)
    :param int export_chunk_id: ID of the chunk (Tenable IO)
    :rtype: None
    """
    # create tenable io directory
    data_dir.mkdir(parents=True, exist_ok=True)
    fn = data_dir / f"{export_type}-{export_uuid}-{export_chunk_id}.json"
    # append file to path
    with open(file=fn, mode="w", encoding="utf-8") as file_object:
        json.dump(data, file_object)


def process_to_regscale(data_dir: Path, ssp_id: int, existing_assets: List[Asset]) -> None:
    """
    Process the data from Tenable IO to RegScale

    :param Path data_dir: The directory where the data is stored
    :param int ssp_id: The ID of the System Security Plan in RegScale
    :param List[Asset] existing_assets: List of existing assets in RegScale
    :rtype: None
    :returns None
    """
    # get all files in the directory
    files = list(data_dir.glob("*.json"))
    if not files:
        logger.warning("No Tenable files found in %s.", data_dir)
        return
    logger.info("Processing %i chunked file(s) from Tenable...", len(list(files)))
    for file in files:
        logger.info("Processing chunked data: %s", file)
        file_assets = []
        with open(file=file, mode="r", encoding="utf-8") as file_object:
            tenable_io_data = json.load(file_object)
            for asset in tenable_io_data:
                file_assets.append(TenableIOAsset(**asset))
        TenableIOAsset.sync_to_regscale(assets=file_assets, ssp_id=ssp_id, existing_assets=existing_assets)
        # remove processed file
        file.unlink()
